@@ -8,14 +8,14 @@ from sql import Table, Literal, Null
 from sql.functions import Overlay, Position
 from sql.aggregate import Count
 from sql.operators import Concat
-from sql.conditionals import Case
 
-from trytond.model import Workflow, ModelView, ModelSQL, fields
+from trytond.model import Workflow, ModelView, ModelSQL, fields, \
+    sequence_ordered
 from trytond.modules.company import CompanyReport
 from trytond.wizard import Wizard, StateAction, StateView, StateTransition, \
     Button
 from trytond import backend
-from trytond.pyson import Eval, Bool, If, PYSONEncoder, Id
+from trytond.pyson import Eval, Bool, If, PYSONEncoder
 from trytond.transaction import Transaction
 from trytond.pool import Pool
 
@@ -33,6 +33,14 @@ _STATES = {
     }
 _DEPENDS = ['state']
 _ZERO = Decimal(0)
+STATES = [
+    ('draft', 'Draft'),
+    ('quotation', 'Quotation'),
+    ('confirmed', 'Confirmed'),
+    ('processing', 'Processing'),
+    ('done', 'Done'),
+    ('cancel', 'Canceled'),
+    ]
 
 
 class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
@@ -52,14 +60,7 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
     reference = fields.Char('Reference', select=True)
     description = fields.Char('Description', size=None, states=_STATES,
         depends=_DEPENDS)
-    state = fields.Selection([
-        ('draft', 'Draft'),
-        ('quotation', 'Quotation'),
-        ('confirmed', 'Confirmed'),
-        ('processing', 'Processing'),
-        ('done', 'Done'),
-        ('cancel', 'Canceled'),
-    ], 'State', readonly=True, required=True)
+    state = fields.Selection(STATES, 'State', readonly=True, required=True)
     purchase_date = fields.Date('Purchase Date',
         states={
             'readonly': ~Eval('state').in_(['draft', 'quotation']),
@@ -69,7 +70,6 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
     payment_term = fields.Many2One('account.invoice.payment_term',
         'Payment Term', states={
             'readonly': ~Eval('state').in_(['draft', 'quotation']),
-            'required': ~Eval('state').in_(['draft', 'quotation', 'cancel']),
             },
         depends=['state'])
     party = fields.Many2One('party.party', 'Party', required=True,
@@ -139,6 +139,14 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
     invoices_recreated = fields.Many2Many(
             'purchase.purchase-recreated-account.invoice',
             'purchase', 'invoice', 'Recreated Invoices', readonly=True)
+    delivery_date = fields.Date(
+        "Delivery Date",
+        states={
+            'readonly': Eval('state').in_([
+                    'processing', 'done', 'cancel']),
+            },
+        depends=['state'],
+        help="The default delivery date for each line.")
     shipment_state = fields.Selection([
             ('none', 'None'),
             ('waiting', 'Waiting'),
@@ -191,7 +199,6 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
                 'quote': {
                     'pre_validate': [
                         ('purchase_date', '!=', None),
-                        ('payment_term', '!=', None),
                         ('invoice_address', '!=', None),
                         ],
                     'invisible': Eval('state') != 'draft',
@@ -206,14 +213,10 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
                 'handle_invoice_exception': {
                     'invisible': ((Eval('invoice_state') != 'exception')
                         | (Eval('state') == 'cancel')),
-                    'readonly': ~Eval('groups', []).contains(
-                        Id('purchase', 'group_purchase')),
                     },
                 'handle_shipment_exception': {
                     'invisible': ((Eval('shipment_state') != 'exception')
                         | (Eval('state') == 'cancel')),
-                    'readonly': ~Eval('groups', []).contains(
-                        Id('purchase', 'group_purchase')),
                     },
                 })
         # The states where amounts are cached
@@ -572,8 +575,11 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
                     })
 
     def get_rec_name(self, name):
-        return (self.number or str(self.id)
-            + ' - ' + self.party.name)
+        if self.number:
+            return self.number
+        elif self.reference:
+            return '[%s]' % self.reference
+        return '(%s)' % self.id
 
     @classmethod
     def search_rec_name(cls, name, clause):
@@ -582,13 +588,10 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
             bool_op = 'AND'
         else:
             bool_op = 'OR'
-        names = value.split(' - ', 1)
         domain = [bool_op,
-            ('number', operator, names[0]),
-            ('reference', operator, names[0]),
+            ('number', operator, value),
+            ('reference', operator, value),
             ]
-        if len(names) != 1 and names[1]:
-            domain = [bool_op, domain, ('party', operator, names[1])]
         return domain
 
     @classmethod
@@ -687,18 +690,17 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
             self.raise_user_error('missing_account_payable',
                     error_args=(self.party.rec_name,))
 
-        invoice_lines = {}
+        invoice_lines = []
         for line in self.lines:
-            ilines = line.get_invoice_line()
-            if ilines:
-                invoice_lines[line.id] = ilines
+            invoice_lines.append(line.get_invoice_line())
+        invoice_lines = list(chain(*invoice_lines))
         if not invoice_lines:
             return
 
         invoice = self._get_invoice_purchase()
-        invoice.lines = list(chain.from_iterable(
-                invoice_lines[l.id] for l in self.lines
-                if l.id in invoice_lines))
+        if getattr(invoice, 'lines', None):
+            invoice_lines = list(invoice.lines) + invoice_lines
+        invoice.lines = invoice_lines
         invoice.save()
 
         Invoice.update_taxes([invoice])
@@ -852,47 +854,55 @@ class PurchaseRecreadtedInvoice(ModelSQL):
             ondelete='RESTRICT', select=True, required=True)
 
 
-class PurchaseLine(ModelSQL, ModelView):
+class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
     'Purchase Line'
     __name__ = 'purchase.line'
     _rec_name = 'description'
     purchase = fields.Many2One('purchase.purchase', 'Purchase',
-            ondelete='CASCADE', select=True, required=True)
-    sequence = fields.Integer('Sequence')
+        ondelete='CASCADE', select=True, required=True,
+        states={
+            'readonly': ((Eval('purchase_state') != 'draft')
+                & Bool(Eval('purchase'))),
+            },
+        depends=['purchase_state'])
     type = fields.Selection([
         ('line', 'Line'),
         ('subtotal', 'Subtotal'),
         ('title', 'Title'),
         ('comment', 'Comment'),
-        ], 'Type', select=True, required=True)
+        ], 'Type', select=True, required=True,
+        states={
+            'readonly': Eval('purchase_state') != 'draft',
+            },
+        depends=['purchase_state'])
     quantity = fields.Float('Quantity',
         digits=(16, Eval('unit_digits', 2)),
         states={
             'invisible': Eval('type') != 'line',
             'required': Eval('type') == 'line',
-            'readonly': ~Eval('_parent_purchase'),
+            'readonly': Eval('purchase_state') != 'draft',
             },
-        depends=['unit_digits', 'type'])
+        depends=['unit_digits', 'type', 'purchase_state'])
     unit = fields.Many2One('product.uom', 'Unit',
         ondelete='RESTRICT',
         states={
             'required': Bool(Eval('product')),
             'invisible': Eval('type') != 'line',
-            'readonly': ~Eval('_parent_purchase'),
+            'readonly': Eval('purchase_state') != 'draft',
             },
         domain=[
             If(Bool(Eval('product_uom_category')),
                 ('category', '=', Eval('product_uom_category')),
                 ('category', '!=', -1)),
             ],
-        depends=['product', 'type', 'product_uom_category'])
+        depends=['product', 'type', 'product_uom_category', 'purchase_state'])
     unit_digits = fields.Function(fields.Integer('Unit Digits'),
         'on_change_with_unit_digits')
     product = fields.Many2One('product.product', 'Product',
         ondelete='RESTRICT', domain=[('purchasable', '=', True)],
         states={
             'invisible': Eval('type') != 'line',
-            'readonly': ~Eval('_parent_purchase'),
+            'readonly': Eval('purchase_state') != 'draft',
             },
         context={
             'locations': If(Bool(Eval('_parent_purchase', {}).get(
@@ -902,7 +912,14 @@ class PurchaseLine(ModelSQL, ModelView):
             'stock_date_end': Eval('_parent_purchase', {}).get(
                 'purchase_date'),
             'stock_skip_warehouse': True,
-            }, depends=['type'])
+            # From _get_context_purchase_price
+            'currency': Eval('_parent_purchase', {}).get('currency'),
+            'supplier': Eval('_parent_purchase', {}).get('party'),
+            'purchase_date': Eval('_parent_purchase', {}).get('purchase_date'),
+            'uom': Eval('unit'),
+            'taxes': Eval('taxes', []),
+            'quantity': Eval('quantity'),
+            }, depends=['type', 'purchase_state'])
     product_uom_category = fields.Function(
         fields.Many2One('product.uom.category', 'Product Uom Category'),
         'on_change_with_product_uom_category')
@@ -910,16 +927,20 @@ class PurchaseLine(ModelSQL, ModelView):
         states={
             'invisible': Eval('type') != 'line',
             'required': Eval('type') == 'line',
-            }, depends=['type'])
+            'readonly': Eval('purchase_state') != 'draft',
+            }, depends=['type', 'purchase_state'])
     amount = fields.Function(fields.Numeric('Amount',
             digits=(16,
                 Eval('_parent_purchase', {}).get('currency_digits', 2)),
             states={
                 'invisible': ~Eval('type').in_(['line', 'subtotal']),
-                'readonly': ~Eval('_parent_purchase'),
                 },
             depends=['type']), 'get_amount')
-    description = fields.Text('Description', size=None, required=True)
+    description = fields.Text('Description', size=None, required=True,
+        states={
+            'readonly': Eval('purchase_state') != 'draft',
+            },
+        depends=['purchase_state'])
     note = fields.Text('Note')
     taxes = fields.Many2Many('purchase.line-account.tax',
         'line', 'tax', 'Taxes',
@@ -932,7 +953,8 @@ class PurchaseLine(ModelSQL, ModelView):
             ],
         states={
             'invisible': Eval('type') != 'line',
-            }, depends=['type'])
+            'readonly': Eval('purchase_state') != 'draft',
+            }, depends=['type', 'purchase_state'])
     invoice_lines = fields.One2Many('account.invoice.line', 'origin',
         'Invoice Lines', readonly=True)
     moves = fields.One2Many('stock.move', 'origin', 'Moves', readonly=True)
@@ -949,22 +971,43 @@ class PurchaseLine(ModelSQL, ModelView):
             'To Location'), 'get_to_location')
     delivery_date = fields.Function(fields.Date('Delivery Date',
             states={
-                'invisible': Eval('type') != 'line',
+                'invisible': ((Eval('type') != 'line')
+                    | Eval('delivery_date_edit', False)),
                 },
-            depends=['type']),
+            depends=['type', 'delivery_date_edit']),
         'on_change_with_delivery_date')
+    delivery_date_edit = fields.Boolean(
+        "Edit Delivery Date",
+        states={
+            'invisible': Eval('type') != 'line',
+            'readonly': Eval('purchase_state').in_([
+                    'processing', 'done', 'cancel']),
+            },
+        depends=['type', 'purchase_state'],
+        help="Check to edit the delivery date.")
+    delivery_date_store = fields.Date(
+        "Delivery Date",
+        states={
+            'invisible': ((Eval('type') != 'line')
+                | ~Eval('delivery_date_edit', False)),
+            'readonly': Eval('purchase_state').in_([
+                    'processing', 'done', 'cancel']),
+            },
+        depends=['type', 'delivery_date_edit', 'purchase_state'])
+    purchase_state = fields.Function(
+        fields.Selection(STATES, 'Purchase State'),
+        'on_change_with_purchase_state')
 
     @classmethod
     def __setup__(cls):
         super(PurchaseLine, cls).__setup__()
-        cls._order.insert(0, ('sequence', 'ASC'))
         cls._error_messages.update({
                 'supplier_location_required': ('Purchase "%(purchase)s" '
                     'misses the supplier location for line "%(line)s".'),
                 'missing_account_expense': ('Product "%(product)s" of '
                     'purchase %(purchase)s misses an expense account.'),
-                'missing_account_expense_property': ('Purchase "%(purchase)s" '
-                    'misses an "account expense" default property.'),
+                'missing_default_account_expense': ('Purchase "%(purchase)s" '
+                    'misses a default "account expense".'),
                 })
 
     @classmethod
@@ -985,35 +1028,36 @@ class PurchaseLine(ModelSQL, ModelView):
         # Migration from 2.4: drop required on sequence
         table.not_null_action('sequence', action='remove')
 
-    @staticmethod
-    def order_sequence(tables):
-        table, _ = tables[None]
-        return [Case((table.sequence == Null, 0), else_=1), table.sequence]
-
     @staticmethod
     def default_type():
         return 'line'
 
-    def get_move_done(self, name):
-        Uom = Pool().get('product.uom')
-        done = True
+    @classmethod
+    def default_delivery_date_edit(cls):
+        return False
+
+    @property
+    def _move_remaining_quantity(self):
+        "Compute the remaining quantity to receive"
+        pool = Pool()
+        Uom = pool.get('product.uom')
         if self.type != 'line' or not self.product:
-            return True
+            return
         if self.product.type == 'service':
-            return True
-        skip_ids = set(x.id for x in self.moves_recreated
-            + self.moves_ignored)
-        quantity = self.quantity
+            return
+        skip_ids = set(x.id for x in self.moves_ignored)
+        quantity = abs(self.quantity)
         for move in self.moves:
-            if move.state != 'done' \
-                    and move.id not in skip_ids:
-                done = False
-                break
-            quantity -= Uom.compute_qty(move.uom, move.quantity, self.unit)
-        if done:
-            if self.unit.round(quantity) > 0.0:
-                done = False
-        return done
+            if move.state == 'done' or move.id in skip_ids:
+                quantity -= Uom.compute_qty(move.uom, move.quantity, self.unit)
+        return quantity
+
+    def get_move_done(self, name):
+        quantity = self._move_remaining_quantity
+        if quantity is None:
+            return True
+        else:
+            return self.unit.round(quantity) <= 0
 
     def get_move_exception(self, name):
         skip_ids = set(x.id for x in self.moves_ignored
@@ -1078,7 +1122,7 @@ class PurchaseLine(ModelSQL, ModelView):
                 if tax_ids:
                     taxes.extend(tax_ids)
                 continue
-            taxes.append(tax)
+            taxes.append(tax.id)
         if party and party.supplier_tax_rule:
             tax_ids = party.supplier_tax_rule.apply(None, pattern)
             if tax_ids:
@@ -1086,7 +1130,7 @@ class PurchaseLine(ModelSQL, ModelView):
         self.taxes = taxes
 
         category = self.product.purchase_uom.category
-        if not self.unit or self.unit not in category.uoms:
+        if not self.unit or self.unit.category != category:
             self.unit = self.product.purchase_uom
             self.unit_digits = self.product.purchase_uom.digits
 
@@ -1175,7 +1219,9 @@ class PurchaseLine(ModelSQL, ModelView):
             return self.purchase.party.supplier_location.id
 
     @fields.depends('product', 'quantity', 'moves',
-        '_parent_purchase.purchase_date', '_parent_purchase.party')
+        '_parent_purchase.purchase_date', '_parent_purchase.party',
+        'delivery_date_edit', 'delivery_date_store',
+        '_parent_purchase.delivery_date')
     def on_change_with_delivery_date(self, name=None):
         if self.moves:
             dates = filter(
@@ -1184,6 +1230,10 @@ class PurchaseLine(ModelSQL, ModelView):
                 return min(dates)
             else:
                 return
+        if self.delivery_date_edit:
+            return self.delivery_date_store
+        if self.purchase and self.purchase.delivery_date:
+            return self.purchase.delivery_date
         if (self.product
                 and self.quantity is not None
                 and self.quantity > 0
@@ -1198,11 +1248,17 @@ class PurchaseLine(ModelSQL, ModelView):
                         return None
                     return delivery_date
 
+    @fields.depends('purchase', '_parent_purchase.state')
+    def on_change_with_purchase_state(self, name=None):
+        if self.purchase:
+            return self.purchase.state
+
     def get_invoice_line(self):
         'Return a list of invoice line for purchase line'
         pool = Pool()
-        Property = pool.get('ir.property')
         InvoiceLine = pool.get('account.invoice.line')
+        AccountConfiguration = pool.get('account.configuration')
+        account_config = AccountConfiguration(1)
 
         invoice_line = InvoiceLine()
         invoice_line.type = self.type
@@ -1238,12 +1294,13 @@ class PurchaseLine(ModelSQL, ModelView):
                         'purchase': self.purchase.rec_name,
                         })
         else:
-            for model in ('product.template', 'product.category'):
-                invoice_line.account = Property.get('account_expense', model)
+            for name in ['default_product_account_expense',
+                    'default_category_account_expense']:
+                invoice_line.account = account_config.get_multivalue(name)
                 if invoice_line.account:
                     break
             if not invoice_line.account:
-                self.raise_user_error('missing_account_expense_property',
+                self.raise_user_error('missing_default_account_expense',
                     {'purchase': self.purchase.rec_name})
         invoice_line.stock_moves = self._get_invoice_line_moves()
         return [invoice_line]
@@ -1368,7 +1425,11 @@ class PurchaseLine(ModelSQL, ModelView):
     def view_attributes(cls):
         return [
             ('/form//field[@name="note"]|/form//field[@name="description"]',
-                'spell', Eval('_parent_purchase', {}).get('party_lang'))]
+                'spell', Eval('_parent_purchase', {}).get('party_lang')),
+            ('//label[@id="delivery_date"]', 'states', {
+                    'invisible': Eval('type') != 'line',
+                    }),
+            ]
 
     @classmethod
     def copy(cls, lines, default=None):
@@ -1416,6 +1477,11 @@ class PurchaseLineRecreatedMove(ModelSQL):
 class PurchaseReport(CompanyReport):
     __name__ = 'purchase.purchase'
 
+    @classmethod
+    def execute(cls, ids, data):
+        with Transaction().set_context(address_with_party=True):
+            return super(PurchaseReport, cls).execute(ids, data)
+
 
 class OpenSupplier(Wizard):
     'Open Suppliers'
@@ -1566,19 +1632,13 @@ class HandleInvoiceException(Wizard):
 
     def transition_handle(self):
         Purchase = Pool().get('purchase.purchase')
-        to_recreate = self.ask.recreate_invoices
-        domain_invoices = self.ask.domain_invoices
 
         purchase = Purchase(Transaction().context['active_id'])
 
-        skip = set(purchase.invoices_ignored)
-        skip.update(purchase.invoices_recreated)
         invoices_ignored = []
         invoices_recreated = []
-        for invoice in purchase.invoices:
-            if invoice not in domain_invoices or invoice in skip:
-                continue
-            if invoice in to_recreate:
+        for invoice in self.ask.domain_invoices:
+            if invoice in self.ask.recreate_invoices:
                 invoices_recreated.append(invoice.id)
             else:
                 invoices_ignored.append(invoice.id)
